"""
===========================================
COBalD - the Opportunistic Balancing Daemon
===========================================

.. image:: https://readthedocs.org/projects/cobald/badge/?version=latest
    :target: http://cobald.readthedocs.io/en/latest/?badge=latest
    :alt: Documentation Status

This is a **draft** for a feedback based balancing system for providing
opportunistic resources.
"""
__title__ = "cobald"
__summary__ = "COBalD - the Opportunistic Balancing Daemon"
__url__ = "https://github.com/MatterMiners/cobald"

__version__ = "0.11.0"
__author__ = "Eileen Kuehn, Max Fischer"
__email__ = "mainekuehn@gmail.com"
__copyright__ = "2018 - 2019 %s" % __author__
__keywords__ = "opportunistic scheduling scheduler demand feedback-loop cobald"
